#include "midi.hpp"
#include "midic.hpp"
#include <algorithm>
#include <iostream>

namespace MIDI
{

//--------------------------------------------------------------
// Structures
//--------------------------------------------------------------
struct MIDIC::Event {
	unsigned int lastClock;
	unsigned int lastMessage;
};

//--------------------------------------------------------------
// Constants
//--------------------------------------------------------------
static const unsigned int eventSize[] = {
	2, /* 0x80 NOTE_OFF      */
	2, /* 0x90 NOTE_ON       */
	2, /* 0xA0 POLYKEY_PRESS */
	2, /* 0xB0 CC            */
	1, /* 0xC0 PC            */
	1, /* 0xD0 CHANNEL_PRESS */
	2, /* 0xE0 PITCHBEND     */
	2, /* 0xF0 META EVENT    */
};

//--------------------------------------------------------------
// Prototypes
//--------------------------------------------------------------


//==============================================================
// Functions
//==============================================================
unsigned short 
big2little_s(unsigned short d)
{
	unsigned char l, m;
	l = static_cast<unsigned char>(d >> 8);
	m = static_cast<unsigned char>(d & 0xff);
	return (m << 8) + l;
}

unsigned int 
big2little_i(unsigned int d)
{
	unsigned short l, m;
	l = big2little_s(static_cast<unsigned short>(d >> 16));
	m = big2little_s(static_cast<unsigned short>(d & 0xffff));
	return (m << 16) + l;
}

void 
MIDIC::dmid_append(
	const unsigned int clock, 
	const int mes, 
	const int lsb, 
	const int msb )
{
	unsigned char m = mes & 0xf0;
	unsigned char ch = mes & 0x0f;

	DMIDI_EVENT ev;

	if (m == MIDI::PITCHBEND) {
		struct DMIDI_EVENT_SHORT {
			unsigned int clock;
			unsigned char ch;
			unsigned char mes;
			unsigned short data;
		};
		DMIDI_EVENT_SHORT& e = reinterpret_cast<DMIDI_EVENT_SHORT&>(ev);
		short data = static_cast<short>(lsb + (msb << 7)) - 8192;
		e.clock = clock;
		e.ch    = ch;
		e.mes   = m;
		e.data  = data;
	} else {
		ev.clock = clock;
		ev.ch    = ch;
		ev.mes   = m;
		ev.lsb   = lsb;
		ev.msb   = msb;
	}
	dmidi->events.push_back(ev);

	return;
}

static bool 
LessClock(const DMIDI_EVENT& lhs, const DMIDI_EVENT& rhs)
{
	return lhs.clock < rhs.clock;
}

DMIDI *
MIDIC::convert_smf(char* smf_data, unsigned int tpqn)
{
	HEADER_CHUNK* h_chunk = (HEADER_CHUNK *)smf_data;
	unsigned char* pData = (unsigned char *)smf_data;
	unsigned int smf_format;
	unsigned int smf_track;
	unsigned int smf_time;
	unsigned int raw_tpqn;

	convertion_tpqn = tpqn;

	std::cout << "Converting SMF..." << std::endl;

	// Check Header
	if (h_chunk->type[0] != 'M' &&
		h_chunk->type[1] != 'T' &&
		h_chunk->type[2] != 'h' &&
		h_chunk->type[3] != 'd') {
		std::cerr << "MIDIC: HeaderChunk Illegal, abort..." << std::endl;
		return NULL;
	}
	if (h_chunk->length != 0x06000000) {
		std::cerr << "MIDIC: HeaderChunk Length Illegal, abort..." << std::endl;
		return NULL;
	}

	smf_format = big2little_s(h_chunk->format);
	smf_track = big2little_s(h_chunk->track);
	smf_time = big2little_s(h_chunk->time);
	raw_tpqn = (smf_time & 0x7fff);

	pData += sizeof(HEADER_CHUNK);

	dmidi = new DMIDI;
	dmidi->tpqn = raw_tpqn;

	for (unsigned int track = 0; track < smf_track; track++) {
		TRACK_CHUNK *t_chunk = reinterpret_cast<TRACK_CHUNK *>(pData);

		// Check Header
		if (t_chunk->type[0] != 'M' &&
			t_chunk->type[1] != 'T' &&
			t_chunk->type[2] != 'r' &&
			t_chunk->type[3] != 'k') {
			std::cerr << "MIDIC: TrackChunk Illegal, abort..." << std::endl;
			delete dmidi;
			return NULL;
		}

		unsigned int length = big2little_i(t_chunk->length);
		unsigned char* p = reinterpret_cast<unsigned char*>(&t_chunk->data[0]);
		unsigned char* pend = p + length;
		Event event = { 0 };

		while (p < pend) {
			p = parse_deltatime(p, &event);
			p = parse_event(p, &event);
		}
		pData = p;
	}
	
	// \[g
	std::stable_sort(dmidi->events.begin(), dmidi->events.end(), LessClock);
	// I[}[J[t
	unsigned int last_clock = dmidi->events.back().clock;
	dmid_append(last_clock, 0xff, 0xff, 0xff);

	dmidi->tpqn = convertion_tpqn;

	return dmidi;
}


unsigned char* 
MIDIC::parse_deltatime(unsigned char* p, Event* event)
{
	int len = 0;
	do {
		len <<= 7;
		len += (*p & 0x7f);
	} while ((*p++ & 0x80) == 0x80);
	event->lastClock += len;
	return p;
}

unsigned char* 
MIDIC::parse_event(unsigned char* p, Event* event)
{
	unsigned char mes = *p;
	unsigned int clock = event->lastClock * convertion_tpqn / dmidi->tpqn;

	// META EVENT or not?
	if ((mes & 0xf0) == 0xf0) {
		if (mes == MIDI::META) {
			if (p[1] == 0x51) { 
				// META EVENT : BPM Change
				int arg = (p[3] << 16) + (p[4] << 8) + p[5];
				arg = static_cast<int>(60.0L / (static_cast<double>(arg) / 1000000.0L));
//				printf("tempo : %d\n", arg);
				dmid_append(
					clock,
					DMIDI_SYS_BPM,
					arg & 0xff,
					arg >> 8
				);
			}
			p += p[2] + 3;
		} else {
			// SYSTEM EXCLUSIVE MESSAGE
			int len = 0;
			int pp = 1;
			do {
				len <<= 7;
				len += p[pp] & 0x7f;
				pp++;
			} while ( p[pp - 1] & 0x80 );
			p += len + pp;
		}
		return p;
	}

	if (mes < 0x80) {
		mes = event->lastMessage;
	} else {
		event->lastMessage = mes;
		p++;
	}

	{
		int size = eventSize[((mes & 0xf0) / 16)-8];
		int lsb = p[0];
		int msb = 0;
		if (size == 2) {
			msb = p[1];
		}
		dmid_append(clock, mes, lsb, msb);
		p += size;
	}

	return p;
}

DMIDI* 
MIDIC::convert_clock_to_frame(DMIDI* dmidi_data, unsigned int fps)
{
	std::vector<DMIDI_EVENT>::iterator event, event_end;
	event_end = dmidi_data->events.end();

	unsigned int last_clock = 0;
	unsigned int last_time = 0;
	unsigned int tempo = 120;
	unsigned int tpqn = dmidi_data->tpqn;

	// convert tick to time[ms]
	for (event = dmidi_data->events.begin(); event != event_end; ++event)
	{
		unsigned int clock = event->clock - last_clock;
		unsigned int tick = clock * 60000 / (tempo * tpqn);
		unsigned int time = last_time + tick;

		if ((event->mes & 0xf0) == DMIDI_SYS 
			&& (event->ch | DMIDI_SYS) == DMIDI_SYS_BPM)
		{
			tempo = event->lsb + (event->msb << 8);
			last_clock = event->clock;
			last_time = time;
		}
		event->clock = time;
	}

	event--;
	{
		int sec = event->clock / 1000 / 60;
		int min = event->clock / 1000 % 60;
/*		printf("   %02d:%02d, %02x %02x %02x %02x\n", 
			sec, min,
			event->ch,
			event->mes,
			event->lsb,
			event->msb
		);*/
	}

	// convet time[ms] to frame
	for (event = dmidi_data->events.begin(); event != event_end; ++event)
	{
		event->clock = event->clock * fps / 1000;
	}
	event--;
/*	printf("%8d, %02x %02x %02x %02x\n", 
		event->clock,
		event->ch,
		event->mes,
		event->lsb,
		event->msb
	);*/


	return dmidi_data;
}

} // namespace MIDI
